@@ -1,11 +1,9 @@
 # Copyright (C) 2008 John Paulett (john -at- paulett.org)
-# Copyright (C) 2009-2021 David Aguilar (davvid -at- gmail.com)
+# Copyright (C) 2009-2024 David Aguilar (davvid -at- gmail.com)
 # All rights reserved.
 #
 # This software is licensed as described in the file COPYING, which
 # you should have received as part of this distribution.
-from __future__ import absolute_import, division, unicode_literals
-
 import collections
 import os
 import unittest
@@ -38,7 +36,7 @@ class Thing:
 
     def __iter__(self):
         for attr in [
-            x for x in getattr(self.__class__, "__dict__") if not x.startswith("__")
+            x for x in getattr(self.__class__, '__dict__') if not x.startswith('__')
         ]:
             yield attr, getattr(self, attr)
 
@@ -51,6 +49,9 @@ class Capture:
         self.args = args
         self.kwargs = kwargs
 
+    def __repr__(self):
+        return object.__repr__(self) + ('(%r, %r)' % (self.args, self.kwargs))
+
 
 class ThingWithProps:
     def _get_identity(self):
@@ -111,7 +112,7 @@ class Outer:
 
 
 class MySlots:
-    __slots__ = ("alpha", "__beta")
+    __slots__ = ('alpha', '__beta')
 
     def __init__(self):
         self.alpha = 1
@@ -119,7 +120,7 @@ class MySlots:
 
 
 class MyPropertiesSlots:
-    __slots__ = ("alpha", "arr")
+    __slots__ = ('alpha', 'arr')
 
     def __init__(self):
         self.alpha = 1
@@ -131,7 +132,7 @@ class MyPropertiesSlots:
 
     @property
     def other_object(self):
-        return Thing("Test")
+        return Thing('Test')
 
     def __eq__(self, other):
         return (
@@ -153,7 +154,7 @@ class MyPropertiesDict:
 
     @property
     def other_object(self):
-        return Thing("Test")
+        return Thing('Test')
 
     def __eq__(self, other):
         return (
@@ -164,9 +165,18 @@ class MyPropertiesDict:
         )
 
 
+# see issue #478
+class SafeData:
+    __slots__ = ()
+
+
+class SafeString(str, SafeData):
+    __slots__ = ()
+
+
 def on_missing_callback(class_name):
     # not actually a runtime problem but it doesn't matter
-    warnings.warn("The unpickler couldn't find %s" % class_name, RuntimeWarning)
+    warnings.warn('The unpickler could not find %s' % class_name, RuntimeWarning)
 
 
 class PicklingTestCase(unittest.TestCase):
@@ -194,13 +204,13 @@ class PicklingTestCase(unittest.TestCase):
         self.assertEqual({tags.B64: encoded}, self.pickler.flatten(data))
 
     def test_decode_base85(self):
-        pickled = {tags.B85: 'P{Y4;Xv4O{u^=-c'}
-        expected = u'P\u00ffth\u00f6\u00f1 3!'.encode('utf-8')
+        expected = 'Pÿthöñ 3!'.encode('utf-8')
+        pickled = {tags.B85: util.b85encode(expected)}
         self.assertEqual(expected, self.unpickler.restore(pickled))
 
     def test_base85_still_handles_base64(self):
-        pickled = {tags.B64: 'UMO/dGjDtsOxIDMh'}
-        expected = u'P\u00ffth\u00f6\u00f1 3!'.encode('utf-8')
+        expected = 'Pÿthöñ 3!'.encode('utf-8')
+        pickled = {tags.B64: util.b64encode(expected)}
         self.assertEqual(expected, self.unpickler.restore(pickled))
 
     def test_string(self):
@@ -247,6 +257,12 @@ class PicklingTestCase(unittest.TestCase):
         self.assertEqual(listD, self.pickler.flatten(listD))
         self.assertEqual(listD, self.unpickler.restore(listD))
 
+    def test_nonetype(self):
+        typ = type(None)
+        typ_pickled = jsonpickle.encode(typ)
+        typ_unpickled = jsonpickle.decode(typ_pickled)
+        self.assertEqual(typ, typ_unpickled)
+
     def test_set(self):
         setlist = ['orange', 'apple', 'grape']
         setA = set(setlist)
@@ -518,12 +534,12 @@ class PicklingTestCase(unittest.TestCase):
                 jsonpickle.decode(case, on_missing='warn')
                 self.assertTrue(issubclass(w[-1].category, UserWarning))
                 self.assertTrue(
-                    "Unpickler._restore_object could not find" in str(w[-1].message)
+                    'Unpickler._restore_object could not find' in str(w[-1].message)
                 )
 
                 jsonpickle.decode(case, on_missing=on_missing_callback)
                 self.assertTrue(issubclass(w[-1].category, RuntimeWarning))
-                self.assertTrue("The unpickler couldn't find" in str(w[-1].message))
+                self.assertTrue('The unpickler could not find' in str(w[-1].message))
 
             self.assertTrue(
                 jsonpickle.decode(case, on_missing='ignore')
@@ -540,20 +556,20 @@ class PicklingTestCase(unittest.TestCase):
 
     def test_private_slot_members(self):
         obj = jsonpickle.loads(jsonpickle.dumps(MySlots()))
-        alpha = getattr(obj, "alpha", "(missing alpha)")
-        beta = getattr(obj, "_" + obj.__class__.__name__ + "__beta", "(missing beta)")
+        alpha = getattr(obj, 'alpha', '(missing alpha)')
+        beta = getattr(obj, '_' + obj.__class__.__name__ + '__beta', '(missing beta)')
         self.assertEqual(alpha, beta)
 
     def test_include_properties_slots(self):
         obj = MyPropertiesSlots()
         dumped = jsonpickle.dumps(obj, include_properties=True)
-        self.assertTrue("py/property" in dumped)
+        self.assertTrue('py/property' in dumped)
         self.assertEqual(obj, jsonpickle.loads(dumped))
 
     def test_include_properties_dict(self):
         obj = MyPropertiesDict()
         dumped = jsonpickle.dumps(obj, include_properties=True)
-        self.assertTrue("py/property" in dumped)
+        self.assertTrue('py/property' in dumped)
         self.assertEqual(obj, jsonpickle.loads(dumped))
 
     def test_load_non_fully_qualified_classes(self):
@@ -561,10 +577,22 @@ class PicklingTestCase(unittest.TestCase):
         obj = MyPropertiesSlots()
         encoded = jsonpickle.encode(obj)
         # MyPropertiesSlots and MyPropertiesDict have compatible eq methods
-        self.assertEqual(
-            obj,
-            jsonpickle.decode(encoded, classes={"MyPropertiesSlots": MyPropertiesDict}),
+        decoded = jsonpickle.decode(
+            encoded, classes={'MyPropertiesSlots': MyPropertiesDict}
         )
+        self.assertIsInstance(decoded, MyPropertiesDict)
+        self.assertEqual(obj, decoded)
+
+    def test_classes_dict(self):
+        # reuse MyPropertiesSlots because it has a nice eq method
+        obj = MyPropertiesSlots()
+        encoded = jsonpickle.encode(obj)
+        # MyPropertiesSlots and MyPropertiesDict have compatible eq methods
+        decoded = jsonpickle.decode(
+            encoded, classes={MyPropertiesSlots: MyPropertiesDict}
+        )
+        self.assertIsInstance(decoded, MyPropertiesDict)
+        self.assertEqual(obj, decoded)
 
     def test_warnings(self):
         data = os.fdopen(os.pipe()[0])
@@ -990,6 +1018,17 @@ class JSONPickleTestCase(SkippableTest):
         pickler.flatten([liszt, liszt, liszt, liszt, liszt])
         self.assertEqual(pickler._depth, -1)
 
+    def test_readonly_attrs(self):
+        s = SafeString("test")
+        pickled = jsonpickle.encode(s, handle_readonly=True)
+        pickled_json_dict = jsonpickle.backend.json.loads(pickled)
+        # make sure it's giving concise output by erroring if it includes
+        # a default method which is unnecessary
+        self.assertTrue("join" not in pickled_json_dict)
+        unpickled = jsonpickle.decode(pickled)
+        self.assertEqual(unpickled.__class__, SafeString)
+        self.assertEqual(unpickled, s)
+
 
 class PicklableNamedTuple(object):
     """
@@ -1070,6 +1109,11 @@ class PickleProtocol2GetStateDict(PickleProtocol2Thing):
         return {'magic': True}
 
 
+class PickleProtocol2GetStateNestedDict(PickleProtocol2Thing):
+    def __getstate__(self):
+        return {'nested': {'magic': True}}
+
+
 class PickleProtocol2GetStateSlots(PickleProtocol2Thing):
     def __getstate__(self):
         return (None, {'slotmagic': slotmagic})
@@ -1522,6 +1566,20 @@ class PicklingProtocol2TestCase(SkippableTest):
         decoded = jsonpickle.decode(encoded)
         self.assertTrue(decoded.magic)
 
+    def test_restore_nested_dict_state_with_references_preserved(self):
+        """
+        Ensure that nested dicts in "py/state" are not duplicated when tracking
+        objects to preserve references
+        """
+        instance1 = PickleProtocol2GetStateNestedDict('whatevs')
+        instance2 = PickleProtocol2GetStateNestedDict('different')
+        encoded = jsonpickle.encode([instance1, instance1, instance2, instance2])
+        decoded = jsonpickle.decode(encoded)
+        self.assertTrue(decoded[0].nested['magic'])
+        self.assertIs(decoded[1], decoded[0])
+        self.assertTrue(decoded[2].nested['magic'])
+        self.assertIs(decoded[3], decoded[2])
+
     def test_restore_slots_state(self):
         """
         Ensure that if getstate returns a 2-tuple with a dict in the second
